/*
	HivelyTracker WinAMP Plugin

	Written by Peter Mulholland
*/

#define _WIN32_IE 0x0200
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>

#include <commctrl.h>

#include <stdlib.h>
#include <stdio.h>

#include "resource.h"

#include "in2.h"
#include "wa_ipc.h"
#include "hvl_replay.h"

// ------------------------------------------------------------------------------------------------
// Function protos
// ------------------------------------------------------------------------------------------------

void Config(HWND hwndParent);
void About(HWND hwndParent);
void Init();
void Quit();
void GetFileInfo(char *file, char *title, int *length_in_ms);
int InfoBox(char *file, HWND hwndParent);
int IsOurFile(char *fn);

int Play(char *fn);
void Stop();
void Pause();
void UnPause();
int IsPaused();

int GetLength();			// get length in ms
int GetOutputTime();		// returns current output time in ms. (usually returns outMod->GetOutputTime()
void SetOutputTime(int time_in_ms);	// seeks to point in stream (in ms). Usually you signal yoru thread to seek, which seeks and calls outMod->Flush()..

// volume stuff
void SetVolume(int volume);	// from 0 to 255.. usually just call outMod->SetVolume
void SetPan(int pan);	// from -127 to 127.. usually just call outMod->SetPan
void SetEQ(int on, char data[10], int preamp);

DWORD WINAPI DecodeThread(void *b);

// ------------------------------------------------------------------------------------------------
// Variables
// ------------------------------------------------------------------------------------------------

const char name[] = "HivelyTracker Plugin 1.6";

In_Module mod = 
{
	IN_VER,										// In version
	name,
	0, 0,
	"hvl\0Hively Tracker Module (*.hvl)\0"
	"ahx\0AHX Module (*.ahx)\0"					// File extensions
	"\0",
	1,											// is_seekable
	TRUE,										// uses output
	Config,
	About,
	Init,
	Quit,
	GetFileInfo,
	InfoBox,
	IsOurFile,
	Play,
	Pause,
	UnPause,
	IsPaused,
	Stop,
	
	GetLength,
	GetOutputTime,
	SetOutputTime,

	SetVolume,
	SetPan,

	0,0,0,0,0,0,0,0,0, // vis stuff

	0,0,	// DSP stuff
	SetEQ,
	NULL,	// setinfo
	0		// out_mod
};

int RunDecodeThread = 0;
int Paused = 0;
HANDLE hDecodeThread = INVALID_HANDLE_VALUE;
struct hvl_tune *ht = NULL;

// DONT CHANGE THESE TWO! Hively only generates 16-bit stereo!
#define BPS		16
#define NCH		2

const int SampleRates[6] = { 11025, 22050, 44100, 48000, 88200, 96000 };
int SelectedRate = 2;
int StereoWidth = 2;
BOOL DoSongEnd = FALSE;

int SampleRate = 44100;
BYTE *SampleBuffer = NULL;
DWORD SampleBufLen = 0;

BOOL ChangeSubSong = FALSE;
uint32 SubSongNr = 0;

BOOL NeedSeek = FALSE;
int SeekPos = 0;
int decode_pos_ms;	// yes winamp, you suck ass for using ints for this

char CurrentFile[MAX_PATH];
uint32 SongLen = 0;

char IniPath[MAX_PATH];

// ------------------------------------------------------------------------------------------------

__declspec(dllexport) In_Module * winampGetInModule2()
{
	return &mod;
}

BOOL CALLBACK ConfigDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{
	static const char *SepWidths[] = { "0%", "25%", "50%", "75%", "100%" };
	static int TmpSelectedRate, TmpStereoWidth;
	static BOOL TmpDoSongEnd;

	switch (message)
	{
		case WM_INITDIALOG:
		{
			int i;

			TmpSelectedRate = SelectedRate;
			TmpStereoWidth = StereoWidth;
			TmpDoSongEnd = DoSongEnd;

			// Sample rate drop box
			for (i = 0; i < 6; i++)
			{
				char str[16];
				_snprintf(str, sizeof(str), "%u", SampleRates[i]);
				SendDlgItemMessage(hwndDlg, IDC_SampleRateCombo, CB_ADDSTRING, (WPARAM) 0, (LPARAM) &str);
			}

			SendDlgItemMessage(hwndDlg, IDC_SampleRateCombo, CB_SETCURSEL, (WPARAM) SelectedRate, (LPARAM) 0);

			// Stereo seperation slider
			SendDlgItemMessage(hwndDlg, IDC_StereoSepSlider, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, 4));
			SendDlgItemMessage(hwndDlg, IDC_StereoSepSlider, TBM_SETPOS, (WPARAM) TRUE, (LPARAM) StereoWidth);
			SendDlgItemMessage(hwndDlg, IDC_StereoSepTxt, WM_SETTEXT, 0, (LPARAM) SepWidths[StereoWidth]);

			// Songend checkbox
			SendDlgItemMessage(hwndDlg, IDC_SongEndChk, BM_SETCHECK, (WPARAM) DoSongEnd ? BST_CHECKED : BST_UNCHECKED, (LPARAM) 0);

			return TRUE;
		}

		case WM_HSCROLL:
		{
			TmpStereoWidth = (int) SendDlgItemMessage(hwndDlg, IDC_StereoSepSlider, TBM_GETPOS, 0, 0);
			SendDlgItemMessage(hwndDlg, IDC_StereoSepTxt, WM_SETTEXT, 0, (LPARAM) SepWidths[TmpStereoWidth]);
			return TRUE;
		}

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_SampleRateCombo:
				{
					if (HIWORD(wParam) == CBN_SELENDOK)
					{
						TmpSelectedRate = (int) SendDlgItemMessage(hwndDlg, IDC_SampleRateCombo, CB_GETCURSEL, 0, 0);
					}
					return TRUE;
				}
				case IDC_SongEndChk:
				{
					if (HIWORD(wParam) ==  BN_CLICKED)
					{
						TmpDoSongEnd = (SendDlgItemMessage(hwndDlg, IDC_SongEndChk, BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
					}

					return TRUE;
				}

				case IDOK:
				{
					char buf[32];
					// apply vars
					SelectedRate = TmpSelectedRate;
					StereoWidth = TmpStereoWidth;
					DoSongEnd = TmpDoSongEnd;

					// Save to config
					_snprintf(buf, sizeof(buf), "%u", SelectedRate);
					WritePrivateProfileString("Settings", "SelectedRate", buf, IniPath);
					_snprintf(buf, sizeof(buf), "%u", StereoWidth);
					WritePrivateProfileString("Settings", "StereoWidth", buf, IniPath);
					WritePrivateProfileString("Settings", "DoSongEnd", DoSongEnd ? "1" : "0", IniPath);
				}

				case IDCANCEL:
					EndDialog(hwndDlg, wParam);
					return TRUE;
			}
	}
	return FALSE;
}

void Config(HWND hwndParent)
{
	//MessageBox(hwndParent, "No configuration yet!", name, MB_OK);
	DialogBox(mod.hDllInstance, MAKEINTRESOURCE(IDD_CONFIGDLG), hwndParent, (DLGPROC) ConfigDlgProc);
}

BOOL CALLBACK AboutDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{
	switch (message)
	{
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDOK:
					EndDialog(hwndDlg, wParam);
					return TRUE;
				case IDC_WEBPAGE:
					ShellExecute(NULL, "open", "http://www.hivelytracker.com", NULL, NULL, SW_SHOWNORMAL);
					return TRUE;
			}
	}
	return FALSE;
}

void About(HWND hwndParent)
{
	DialogBox(mod.hDllInstance, MAKEINTRESOURCE(IDD_ABOUTDLG), hwndParent, (DLGPROC) AboutDlgProc);
}

void Init()
{
	char *path;
	hvl_InitReplayer();

	path = (char *) SendMessage(mod.hMainWindow,WM_WA_IPC,(WPARAM) 0,IPC_GETINIDIRECTORY);
	strcpy(IniPath, path);
	strcat(IniPath, "\\HivelyWinamp.ini");

	// Load config
	SelectedRate = GetPrivateProfileInt("Settings", "SelectedRate", 2, IniPath);
	StereoWidth = GetPrivateProfileInt("Settings", "StereoWidth", 2, IniPath);
	DoSongEnd = GetPrivateProfileInt("Settings", "DoSongEnd", 0, IniPath);
}

void Quit()
{
}

void GetFileInfo(char *file, char *title, int *length_in_ms)
{
	if (file)
	{
		if (*file == 0)		// Info about the current file
		{
			if (ht)
			{
				if (title)
				{
					strcpy(title, ht->ht_Name);
				}
				*length_in_ms = SongLen;
			}
		}
		else				// Info about another file!
		{
			struct hvl_tune *inf;
			inf = hvl_LoadTune(file, SampleRate, StereoWidth);
			if (inf)
			{
				if (title)
					strcpy(title, inf->ht_Name);

				*length_in_ms = hvl_GetLen(inf, 0);

				hvl_FreeTune(inf);
			}
		}
	}
}

BOOL CALLBACK InfoDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{
	switch (message)
	{
		case WM_INITDIALOG:
		{
			int i;
			char buf[256];

			// Instrument list
			for (i = 1; i <= ht->ht_InstrumentNr; i++)
			{
				struct hvl_instrument *in = (ht->ht_Instruments) + i;
				_snprintf(buf, sizeof(buf), "%02d: %s", i, in->ins_Name);
				SendDlgItemMessage(hwndDlg, IDC_InstrumentList, LB_ADDSTRING, 0, (LPARAM)(char *) buf);
			}

			// Module type
			if (strncmp(ht->ht_ID, "HVL", 3) == 0)
			{
				strcpy(buf, "HivelyTracker");
			}
			else if (strncmp(ht->ht_ID, "THX", 3) == 0)
			{
				_snprintf(buf, sizeof(buf), "AHX Version %d", ht->ht_ID[3]);
			}
			else
				strcpy(buf, "unknown!?!?!");

			SendDlgItemMessage(hwndDlg, IDC_Type, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			// Misc info
			_snprintf(buf, sizeof(buf), "%d", ht->ht_Channels);
			SendDlgItemMessage(hwndDlg, IDC_Chans, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			_snprintf(buf, sizeof(buf), "%d", ht->ht_Tempo);
			SendDlgItemMessage(hwndDlg, IDC_Speed, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			_snprintf(buf, sizeof(buf), "%d", ht->ht_PositionNr);
			SendDlgItemMessage(hwndDlg, IDC_Positions, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			// Subsong slider and indicator
			SendDlgItemMessage(hwndDlg, IDC_SubSongSlider, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, ht->ht_SubsongNr));
			SendDlgItemMessage(hwndDlg, IDC_SubSongSlider, TBM_SETPOS, (WPARAM) TRUE, (LPARAM) ht->ht_SongNum);

			_snprintf(buf, sizeof(buf), "%02d/%02d", ht->ht_SongNum+1, ht->ht_SubsongNr+1);
			SendDlgItemMessage(hwndDlg, IDC_SubSongIndicator, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			//_snprintf(buf, sizeof(buf), "tlen %d\ntnr %d", ht->ht_TrackLength, ht->ht_TrackNr);
			//SendDlgItemMessage(hwndDlg, IDC_dbg, WM_SETTEXT, 0, (LPARAM)(char *) buf);

			return TRUE;
		}

		case WM_HSCROLL:
		{
			uint32 ssng = (uint32) SendDlgItemMessage(hwndDlg, IDC_SubSongSlider, TBM_GETPOS, 0, 0);

			if (ssng != SubSongNr)
			{
				char buf[16];

				SubSongNr = ssng;
				ChangeSubSong = TRUE;

				_snprintf(buf, sizeof(buf), "%02d/%02d", SubSongNr+1, ht->ht_SubsongNr+1);
				SendDlgItemMessage(hwndDlg, IDC_SubSongIndicator, WM_SETTEXT, 0, (LPARAM)(char *) buf);
			}

			return TRUE;
		}

		case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
				case IDOK:
					EndDialog(hwndDlg, wParam);
					return TRUE;
			}
		}
	}
	return FALSE;
}

int InfoBox(char *file, HWND hwndParent)
{
	// Argh, winamp is such an arse... we need to handle info dialog for any file, not just the current one
	// Not doing that yet!
	if (ht && strcmp(CurrentFile, file) == 0)
	{
		DialogBoxParam(mod.hDllInstance, MAKEINTRESOURCE(IDD_INFODLG), hwndParent, (DLGPROC) InfoDlgProc, (LPARAM) (char *) file);
	}

	return 0;
}

int IsOurFile(char *fn)
{
	FILE *f = fopen(fn, "rb");
	BYTE id[4];

	// Quick ID check, rather than load the whole file, just load the first 4 bytes and check those
	if (f)
	{
		fread(&id, 4, 1, f);
		fclose(f);

		// HivelyTracker file
		if ((id[0] == 'H') && (id[1] == 'V') && (id[2] == 'L') && (id[3] <= 1))
		{
			return 1;
		}

		// AHX File
		if ((id[0] == 'T') && (id[1] == 'H') && (id[2] == 'X') && (id[3] < 3))
		{
			return 1;
		}
	}

	return 0;
}

int Play(char *fn)
{
	int maxlatency;
	DWORD thread_id;

	SampleRate = SampleRates[SelectedRate];

	ht = hvl_LoadTune(fn, SampleRate, StereoWidth);

	if (!ht)
	{
		MessageBox(mod.hMainWindow, "Failed to load module file!", name, MB_OK);
		return 1;
	}

	strcpy(CurrentFile, fn);

	SampleBufLen = ((SampleRate*NCH*(BPS/8))/50);
	SampleBuffer = (BYTE *) malloc(SampleBufLen * 2);	// DSP may want more, suck!

	if (!SampleBuffer)
	{
		MessageBox(mod.hMainWindow, "Failed to allocate sample buffer!", name, MB_OK);
		return 1;
	}

	maxlatency = mod.outMod->Open(SampleRate, NCH, BPS, -1,-1);

	if (maxlatency < 0) // error opening device
	{
		return 1;
	}

	mod.SetInfo(ht->ht_Channels, SampleRate/1000, 2, 1);

	// initialize vis stuff
	mod.SAVSAInit(maxlatency, SampleRate);
	mod.VSASetInfo(SampleRate, NCH);
	mod.outMod->SetVolume(-666);

	ChangeSubSong = FALSE;
	SubSongNr = 0;
	SongLen = hvl_GetLen(ht, SubSongNr);
	hvl_InitSubsong(ht, SubSongNr);

	NeedSeek = FALSE;
	decode_pos_ms = 0;

	RunDecodeThread = 1;
	hDecodeThread = (HANDLE) CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) DecodeThread, NULL, 0, &thread_id);
	SetThreadPriority(hDecodeThread, THREAD_PRIORITY_HIGHEST);

	return 0;	// Success!
}
 
void Stop()
{
	if (hDecodeThread != INVALID_HANDLE_VALUE)
	{
		RunDecodeThread = 0;

		if (WaitForSingleObject(hDecodeThread, 1000 * 10) == WAIT_TIMEOUT)
		{
			MessageBox(mod.hMainWindow,"Error asking playback thread to die!", name, MB_OK);
			TerminateThread(hDecodeThread,0);
		}

		CloseHandle(hDecodeThread);
		hDecodeThread = INVALID_HANDLE_VALUE;
	}

	mod.outMod->Close();
	mod.SAVSADeInit();

	if (ht)
	{
		hvl_FreeTune(ht);
		ht = NULL;
		CurrentFile[0] = 0;
	}

	if (SampleBuffer)
	{
		free(SampleBuffer);
		SampleBuffer = NULL;
	}
}

void Pause()
{
	Paused=1;
	mod.outMod->Pause(1);
}

void UnPause()
{
	Paused=0;
	mod.outMod->Pause(0);
}

int IsPaused()
{
	return Paused;
}

int GetLength()
{
	return SongLen;
}

int GetOutputTime()
{
	return decode_pos_ms+(mod.outMod->GetOutputTime()-mod.outMod->GetWrittenTime());
}

void SetOutputTime(int time_in_ms)
{
	NeedSeek = TRUE;
	SeekPos = time_in_ms;
}

void SetVolume(int volume)
{
	mod.outMod->SetVolume(volume);
}

void SetPan(int pan)
{
	mod.outMod->SetPan(pan);
}

void SetEQ(int on, char data[10], int preamp)
{
}

DWORD WINAPI DecodeThread(void *b)
{
	int l;

	while (RunDecodeThread)
	{
		if (ChangeSubSong)
		{
			SongLen = hvl_GetLen(ht, SubSongNr);
			hvl_InitSubsong(ht, SubSongNr);
			ChangeSubSong = FALSE;
			mod.outMod->Flush(0);
			decode_pos_ms = 0;
		}

		if (NeedSeek)
		{
			// do seek in here
			decode_pos_ms = SeekPos-(SeekPos % 1000);
			hvl_Seek(ht, decode_pos_ms);
			mod.outMod->Flush(decode_pos_ms);
			NeedSeek = FALSE;
		}

		if (DoSongEnd ? ht->ht_SongEndReached : 0)
		{
			mod.outMod->CanWrite();
			if (!mod.outMod->IsPlaying())
			{
				// We're done, tell winamp its over
				PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
			}
		}

		if (mod.outMod->CanWrite() >= (SampleBufLen << (mod.dsp_isactive() ? 1 : 0)))
		{
			if (DoSongEnd ? !ht->ht_SongEndReached : 1)
			{
				hvl_DecodeFrame(ht, SampleBuffer, SampleBuffer+(BPS/8), (BPS/8)*NCH);

				//decode_pos_ms += SampleRate/(SampleBufLen/((BPS/8)*NCH));
				decode_pos_ms = mod.outMod->GetWrittenTime();
				mod.SAAddPCMData((char *)SampleBuffer,NCH,BPS,decode_pos_ms);
				mod.VSAAddPCMData((char *)SampleBuffer,NCH,BPS,decode_pos_ms);

				if (mod.dsp_isactive())
					l = mod.dsp_dosamples((short *) SampleBuffer,SampleBufLen/NCH/(BPS/8),BPS,NCH,SampleRate)*(NCH*(BPS/8));
				else
					l = SampleBufLen;

				mod.outMod->Write(SampleBuffer, l);
			}
			else
				Sleep(10);
		}
		else
			Sleep(10);
	}

	return 0;
}
